# Copyright 2025 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mocks server calls for testing
"""

from importlib.metadata import version
import json
import os

from iqm.iqm_client import IQMClient, TokenManager
from mockito import ANY, expect, when
from packaging.version import parse
import pytest
import requests
from requests import HTTPError, Response

from iqm.station_control.client.station_control import StationControlClient
from iqm.station_control.interface import models


@pytest.fixture()
def base_url() -> str:
    # NOTE: You should mock all HTTP requests in the tests, so we do not send out actual HTTP requests here!
    return "https://example.com"


@pytest.fixture(scope="function")
def iqm_client_mock(base_url) -> IQMClient:
    expect(requests, times=1).get(
        f"{base_url}/info/client-libraries",
        headers=ANY,
        timeout=ANY,
    ).thenReturn(mock_supported_client_libraries_response())
    when(requests).get(f"{base_url}/about", headers=ANY).thenReturn(MockJsonResponse(200, {}))

    when(StationControlClient)._check_api_versions().thenReturn(None)
    client = IQMClient(base_url)
    client._token_manager = TokenManager()  # Do not use authentication
    when(client._token_manager).get_bearer_token().thenReturn(None)
    return client


@pytest.fixture
def crystal_5_calibration_set():
    path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
        "tests/iqm_client/resources/crystal_5_calibration_set.json",
    )
    with open(path, "r", encoding="utf-8") as data:
        return json.load(data)

        # return models.ObservationSetData(
        #     observation_set_type="calibration-set",
        #     observation_ids=[obs["observation_id"] for obs in calset],
        # )


@pytest.fixture
def crystal_5_quality_metrics(crystal_5_calibration_set):
    path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
        "tests/iqm_client/resources/crystal_5_quality_metric_set.json",
    )
    with open(path, "r", encoding="utf-8") as data:
        quality_metric_set = json.load(data)

    calibration_set_data = models.ObservationSetData(
        observation_set_type="calibration-set",
        # Extract all observation IDs from the list of observations
        observation_ids=[obs["observation_id"] for obs in crystal_5_calibration_set if "observation_id" in obs],
        # Use metadata from the quality metric set as it's the context
        describes_id=quality_metric_set["describes_id"],
        invalid=quality_metric_set["invalid"],
        dut_label=quality_metric_set["dut_label"],
        observation_set_id=quality_metric_set["observation_set_id"],
        created_timestamp=quality_metric_set["created_timestamp"],
        end_timestamp=quality_metric_set["end_timestamp"],
    )

    # Convert the list of observation dicts to a list of ObservationLite model instances
    observations_models = [models.ObservationLite.model_construct(**obs) for obs in quality_metric_set["observations"]]

    return models.QualityMetrics(
        observation_set_type=quality_metric_set["observation_set_type"],
        observation_ids=quality_metric_set["observation_ids"],
        describes_id=quality_metric_set["describes_id"],
        invalid=quality_metric_set["invalid"],
        dut_label=quality_metric_set["dut_label"],
        observation_set_id=quality_metric_set["observation_set_id"],
        created_timestamp=quality_metric_set["created_timestamp"],
        end_timestamp=quality_metric_set["end_timestamp"],
        calibration_set=calibration_set_data,
        observations=observations_models,
    )


class MockBytesResponse:
    def __init__(self, status_code: int, content: bytes, history: list[Response] | None = None):
        self.status_code = status_code
        self.content = content
        self.media_type = "application/octet-stream"
        self.history = history
        self.url = "https://example.com"

    def raise_for_status(self):
        if 400 <= self.status_code < 600:
            raise HTTPError(f"{self.status_code}", response=self)


class MockJsonResponse:
    def __init__(self, status_code: int, json_data: dict | list[dict], history: list[Response] | None = None):
        self.status_code = status_code
        self.json_data = json_data
        self.history = history
        self.url = "https://example.com"

    @property
    def text(self):
        # NOTE cannot handle UUIDs
        return json.dumps(self.json_data)

    def json(self):
        return self.json_data

    def raise_for_status(self):
        if 400 <= self.status_code < 600:
            raise HTTPError(f"{self.status_code}", response=self)


def mock_supported_client_libraries_response(
    iqm_client_name: str = "iqm-client", max_version: str | None = None, min_version: str | None = None
) -> MockJsonResponse:
    client_version = parse(version("iqm-client"))
    min_version = f"{client_version.major}.0" if min_version is None else min_version
    max_version = f"{client_version.major + 1}.0" if max_version is None else max_version
    return MockJsonResponse(
        200,
        {
            iqm_client_name: {
                "name": iqm_client_name,
                "min": min_version,
                "max": max_version,
            }
        },
    )


def mock_about_response():
    return MockJsonResponse(200, {})



# Copyright 2022-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Qiskit backend for IQM quantum computers."""

from __future__ import annotations

from abc import ABC
import logging
from typing import Final

from iqm.iqm_client import (
    DynamicQuantumArchitecture,
    ObservationFinder,
)
from iqm.qiskit_iqm.iqm_target import IQMTarget
from qiskit.providers import BackendV2
from qiskit.transpiler import Target

IQM_TO_QISKIT_GATE_NAME: Final[dict[str, str]] = {"prx": "r", "cz": "cz"}
logger = logging.getLogger(__name__)


class IQMBackendBase(BackendV2, ABC):
    """Abstract base class for various IQM-specific backends.

    Args:
        architecture: Dynamic quantum architecture associated with the backend instance.
        metrics: Optional calibration data and related quality metrics for the transpilation target.
        name: Optional name for the backend instance.

    """

    def __init__(
        self,
        architecture: DynamicQuantumArchitecture,
        *,
        metrics: ObservationFinder | None = None,
        name: str = "IQMBackend",
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self.architecture = architecture
        self.metrics = metrics
        # Qiskit uses integer indices to refer to qubits, so we need to map component names to indices.
        # Because of the way the Target and the transpiler interact, the resonators need to have higher indices than
        # qubits, or else transpiling with optimization_level=0 will fail because of lacking resonator indices.
        qb_to_idx = {qb: idx for idx, qb in enumerate(architecture.qubits + architecture.computational_resonators)}

        # NOTE: both targets include fictional CZs
        self._target = IQMTarget(
            architecture=architecture,
            component_to_idx=qb_to_idx,
            include_resonators=False,
            include_fictional_czs=True,
            metrics=metrics,
        )

        self._target_with_resonators = (
            IQMTarget(
                architecture=architecture,
                component_to_idx=qb_to_idx,
                include_resonators=True,
                include_fictional_czs=True,
                metrics=metrics,
            )
            if "move" in architecture.gates
            else None
        )

        self._qb_to_idx = qb_to_idx
        self._idx_to_qb = {v: k for k, v in qb_to_idx.items()}
        self._coupling_map = self.target.build_coupling_map()

    @property
    def target(self) -> Target:
        """Return the target without computational resonators."""
        return self._target

    @property
    def target_with_resonators(self) -> Target:
        """Return the target with MOVE gates and resonators included.

        Raises:
            ValueError: The backend does not have resonators.

        """
        target = self._target_with_resonators
        if target is None:
            raise ValueError("The backend does not have computational resonators.")
        return target

    @property
    def physical_qubits(self) -> list[str]:
        """Return the list of physical qubits in the backend."""
        return list(self._qb_to_idx)

    def has_resonators(self) -> bool:
        """True iff the backend QPU has computational resonators."""
        return bool(self.architecture.computational_resonators)

    def get_real_target(self) -> Target:
        """Return the real physical target of the backend without fictional CZ gates."""
        return IQMTarget(
            architecture=self.architecture,
            component_to_idx=self._qb_to_idx,
            include_resonators=True,
            include_fictional_czs=False,
            metrics=self.metrics,
        )

    def qubit_name_to_index(self, name: str) -> int:
        """Given an IQM-style qubit name, return the corresponding index in the register.

        Args:
            name: IQM-style qubit name ('QB1', 'QB2', etc.)

        Returns:
            Index of the given qubit in the quantum register.

        Raises:
            ValueError: Qubit name cannot be found on the backend.

        """
        if name not in self._qb_to_idx:
            raise ValueError(f"Qubit '{name}' is not found on the backend.")
        return self._qb_to_idx[name]

    def index_to_qubit_name(self, index: int) -> str:
        """Given a quantum register index, return the corresponding IQM-style qubit name.

        Args:
            index: Qubit index in the quantum register.

        Returns:
            Corresponding IQM-style qubit name ('QB1', 'QB2', etc.).

        Raises:
            ValueError: Qubit index cannot be found on the backend.

        """
        if index not in self._idx_to_qb:
            raise ValueError(f"Qubit index {index} is not found on the backend.")
        return self._idx_to_qb[index]

    def get_scheduling_stage_plugin(self) -> str:
        """Return the plugin that should be used for scheduling the circuits on this backend."""
        return "iqm_default_scheduling"



# Copyright 2024 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Naive transpilation for the IQM Star architecture."""

import warnings

from iqm.iqm_client import Circuit as IQMClientCircuit
from iqm.iqm_client.transpile import ExistingMoveHandlingOptions, transpile_insert_moves
import numpy as np
from pydantic_core import ValidationError
from qiskit import QuantumCircuit, transpile
from qiskit.circuit.library import RGate
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.layout import Layout

from .iqm_backend import IQMBackendBase, IQMTarget
from .iqm_move_layout import generate_initial_layout
from .qiskit_to_iqm import deserialize_instructions, serialize_instructions


class IQMNaiveResonatorMoving(TransformationPass):
    """Naive transpilation pass for resonator moving.

    The logic of this pass is deferred to `iqm-client.transpile_insert_moves`.
    This pass is a wrapper that converts the circuit into the IQMClient Circuit format,
    runs the `transpile_insert_moves` function, and then converts the result back to a Qiskit circuit.

    Args:
        target: Transpilation target.
        existing_moves_handling: How to handle existing MOVE gates in the circuit.

    """

    def __init__(
        self,
        target: IQMTarget,
        existing_moves_handling: ExistingMoveHandlingOptions = ExistingMoveHandlingOptions.KEEP,
    ):
        super().__init__()
        self.target = target
        self.architecture = target.iqm_dqa
        self.idx_to_component = target.iqm_idx_to_component
        self.component_to_idx = target.iqm_component_to_idx
        self.existing_moves_handling = existing_moves_handling

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        """Run the pass on a circuit.

        Args:
            dag: DAG to map.

        Returns:
            Mapped ``dag``.

        Raises:
            TranspilerError: The layout is not compatible with the DAG, or if the input gate set is incorrect.

        """
        # TODO: Temporary hack to get the symbolic parameters to work: replace symbols with (inf, idx).
        # Replace symbolic parameters with indices and store the index to symbol mapping.
        symbolic_gates = {}
        symbolic_index = 0.0
        for node in dag.topological_op_nodes():
            # This only works for prx gates because that has two parameters
            # We use one to mark that it is a symbolic gate (np.inf) and the other to store the index.
            if node.name == "r" and not all(isinstance(param, (float, int)) for param in node.op.params):
                symbolic_gates[symbolic_index] = node.op.params
                dag.substitute_node(node, RGate(np.inf, float(symbolic_index)))
                symbolic_index += 1

        circuit = dag_to_circuit(dag)
        if len(circuit) == 0:
            return dag  # Empty circuit, no need to transpile.
        # For some reason, the dag does not contain the layout, so we need to do a bunch of fixing.
        if self.property_set.get("layout"):
            layout = self.property_set["layout"]
        else:
            # Reconstruct the layout from the dag.
            layout = Layout()
            for qreg in dag.qregs:
                layout.add_register(qreg)
            for i, qubit in enumerate(dag.qubits):
                layout.add(qubit, i)

        # Convert the circuit to the IQMClientCircuit format and run the transpiler.
        iqm_circuit = IQMClientCircuit(
            name="Transpiling Circuit",
            instructions=tuple(serialize_instructions(circuit, self.idx_to_component)),
            metadata=None,
        )
        try:
            routed_iqm_circuit = transpile_insert_moves(
                iqm_circuit,
                self.architecture,
                existing_moves=self.existing_moves_handling,
            )
            routed_circuit = deserialize_instructions(
                list(routed_iqm_circuit.instructions), self.component_to_idx, layout
            )
        except ValidationError as e:  # The Circuit without move gates is empty.
            errors = e.errors()
            if (
                len(errors) == 1
                and errors[0]["msg"] == "Value error, Each circuit should have at least one instruction."
            ):
                circ_args = [circuit.num_ancillas, circuit.num_clbits]
                routed_circuit = QuantumCircuit(*layout.get_registers(), *(arg for arg in circ_args if arg > 0))
            else:
                raise e

        # Create the new DAG and make sure that the qubits are properly ordered.
        ordered_qubits = [layout.get_physical_bits()[i] for i in range(len(layout.get_physical_bits()))]
        new_dag = circuit_to_dag(
            routed_circuit,
            qubit_order=ordered_qubits,
            clbit_order=routed_circuit.clbits,
        )

        # Reinsert the symbolic parameters.
        for node in new_dag.topological_op_nodes():
            # This only works for prx gates because that has two parameters
            # We use one to mark that it is a symbolic gate (np.inf) and the other to store the index.
            if node.name == "r" and not np.isfinite(node.op.params[0]):
                new_dag.substitute_node(node, RGate(*symbolic_gates[int(np.round(node.op.params[1]))]))

        # Update the final_layout with the correct bits.
        if "final_layout" in self.property_set:
            inv_layout = layout.get_physical_bits()
            new_final_layout_dict = {
                physical: inv_layout[dag.find_bit(virtual).index]
                for physical, virtual in self.property_set["final_layout"].get_physical_bits().items()
            }
            resonator_dict = {
                phys: inv_layout[new_dag.find_bit(virt).index]
                for phys, virt in inv_layout.items()
                if phys not in self.property_set["final_layout"].get_physical_bits()
            }
            new_final_layout_dict.update(resonator_dict)
            self.property_set["final_layout"] = Layout(new_final_layout_dict)
        else:
            self.property_set["final_layout"] = layout

        return new_dag


def _get_scheduling_method(
    perform_move_routing: bool,
    optimize_single_qubits: bool,
    remove_final_rzs: bool,
    ignore_barriers: bool,
    existing_moves_handling: ExistingMoveHandlingOptions | None,
) -> str:
    """Determine scheduling based on flags."""
    if perform_move_routing:
        if optimize_single_qubits:
            if not remove_final_rzs and ignore_barriers and existing_moves_handling is None:
                raise ValueError(
                    f"Move gate routing not compatible with {optimize_single_qubits=}, "
                    f"{remove_final_rzs=}, and {ignore_barriers=}."
                )
            if not remove_final_rzs:
                scheduling_method = "move_routing_exact_global_phase"
            elif ignore_barriers:
                scheduling_method = "move_routing_rz_optimization_ignores_barriers"
            else:
                scheduling_method = "move_routing"
        else:
            scheduling_method = "only_move_routing"
        if existing_moves_handling is not None:
            if not scheduling_method.endswith("routing"):
                raise ValueError(
                    "Existing Move handling options are not compatible with `remove_final_rzs` and \
                    `ignore_barriers` options."
                )  # No technical reason for this, just hard to maintain all combinations.
            scheduling_method += "_" + existing_moves_handling.value
    elif optimize_single_qubits:
        scheduling_method = "only_rz_optimization"
        if not remove_final_rzs:
            scheduling_method += "_exact_global_phase"
        if ignore_barriers:
            scheduling_method += "_ignore_barriers"
    else:
        scheduling_method = "default"
    return scheduling_method


def transpile_to_IQM(  # noqa: PLR0913
    circuit: QuantumCircuit,
    backend: IQMBackendBase,
    *,
    initial_layout: Layout | dict | list | None = None,
    perform_move_routing: bool = True,
    optimize_single_qubits: bool = True,
    ignore_barriers: bool = False,
    remove_final_rzs: bool = True,
    existing_moves_handling: ExistingMoveHandlingOptions | None = None,
    restrict_to_qubits: list[int] | list[str] | None = None,
    **qiskit_transpiler_kwargs,
) -> QuantumCircuit:
    """Customized transpilation to IQM backends.

    Works with both the Crystal and Star architectures.

    Note: When transpiling a circuit with MOVE gates, you might need to set ``optimization_level`` lower.
    If ``optimization_level`` is set too high, the transpiler might add single qubit gates onto the resonator,
    which is not supported by the IQM Star architectures. If this in undesired, it is best to have the transpiler
    add the MOVE gates automatically, rather than manually adding them to the circuit.

    Args:
        circuit: The circuit to transpile.
        backend: The backend to transpile to.
        initial_layout: The initial layout to use for the transpilation, same as :func:`~qiskit.compiler.transpile`.
        perform_move_routing: Whether to perform MOVE gate routing.
        optimize_single_qubits: Whether to optimize single qubit gates away.
        ignore_barriers: Whether to ignore barriers when optimizing single qubit gates away.
        remove_final_rzs: Whether to remove the final z rotations. It is recommended always to set this to true as
            the final RZ gates do no change the measurement outcomes of the circuit.
        existing_moves_handling: How to handle existing MOVE gates in the circuit, required if the circuit contains
            MOVE gates.
        restrict_to_qubits: Restrict the transpilation to only use these specific physical qubits. Note that you will
            also have to pass this information to :meth:`.IQMBackend.run` using the ``qubit_mapping`` parameter.
        qiskit_transpiler_kwargs: Arguments to be passed to the Qiskit transpiler.

    Returns:
        Transpiled circuit ready for running on the backend.

    """
    circuit_has_moves = circuit.count_ops().get("move", 0) > 0
    # get the target from the backend
    if circuit_has_moves:
        target = backend.target_with_resonators
        if perform_move_routing and existing_moves_handling is None:
            raise ValueError("The circuit contains MOVE gates but existing_moves_handling is not set.")
    else:
        target = backend.target

    if restrict_to_qubits is not None:
        # qubit names to qiskit indices
        restrict_to_qubits = [backend.qubit_name_to_index(q) if isinstance(q, str) else q for q in restrict_to_qubits]
        target = target.restrict_to_qubits(restrict_to_qubits)

    if circuit_has_moves and initial_layout is None:
        # Create a sensible initial layout if none is provided, since
        # the standard Qiskit transpile function does not do this well with resonators.
        real_target = backend.get_real_target()
        if restrict_to_qubits is not None:
            real_target = real_target.restrict_to_qubits(restrict_to_qubits)
        initial_layout = generate_initial_layout(real_target, circuit)

    # Determine which scheduling method to use
    scheduling_method = qiskit_transpiler_kwargs.pop("scheduling_method", None)
    if scheduling_method is None:
        scheduling_method = _get_scheduling_method(
            perform_move_routing=perform_move_routing,
            optimize_single_qubits=optimize_single_qubits,
            remove_final_rzs=remove_final_rzs,
            ignore_barriers=ignore_barriers,
            existing_moves_handling=existing_moves_handling,
        )
    else:
        warnings.warn(
            f"Scheduling method is set to {scheduling_method}, but it is normally used to pass other transpiler "
            + "options, ignoring the `perform_move_routing`, `optimize_single_qubits`, `remove_final_rzs`, "
            + "`ignore_barriers`, and `existing_moves_handling` arguments."
        )
    qiskit_transpiler_kwargs["scheduling_method"] = scheduling_method
    new_circuit = transpile(
        circuit,
        target=target,
        initial_layout=initial_layout,
        **qiskit_transpiler_kwargs,
    )
    return new_circuit



# Copyright 2022-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transpilation target for IQM quantum computers."""

from __future__ import annotations

from collections.abc import Iterable
import logging
from typing import TypeAlias

from iqm.iqm_client import (
    DynamicQuantumArchitecture,
    GateImplementationInfo,
    GateInfo,
    ObservationFinder,
)
from iqm.qiskit_iqm.move_gate import MoveGate
from qiskit.circuit import Delay, Gate, Parameter, Reset
from qiskit.circuit.library import CZGate, IGate, Measure, RGate
from qiskit.providers import QubitProperties
from qiskit.transpiler import InstructionProperties, Target

Locus: TypeAlias = tuple[str, ...]
"""Sequence of QPU component names on which a gate acts."""
LocusIdx: TypeAlias = tuple[int, ...]
"""Sequence of qubit indices on which a gate acts."""


_QISKIT_IQM_GATE_MAP: dict[str, Gate] = {
    "prx": RGate(Parameter("theta"), Parameter("phi")),
    "measure": Measure(),
    "reset": Reset(),
    "cz": CZGate(),
    "move": MoveGate(),
    "id": IGate(),
}
"""Maps IQM native operation names to corresponding Qiskit gate objects."""

logger = logging.getLogger(__name__)


class IQMTarget(Target):
    """Transpilation target for an IQM architecture.

    Contains the mapping of physical qubit name on the device to qubit index in the Target.

    Args:
        architecture: Quantum architecture that defines the target.
        component_to_idx: Mapping from QPU component names to integer indices used by Qiskit to refer to them.
        include_resonators: Whether to include computational resonators (and MOVE gates) in the target,
            if present in ``architecture``.
        include_fictional_czs: Whether to include "fictional" CZs that are not natively supported,
            but could be routed using MOVE.
        metrics: Optional calibration data and related quality metrics to improve the transpilation.

    """

    def __init__(
        self,
        *,
        architecture: DynamicQuantumArchitecture,
        component_to_idx: dict[str, int],
        include_resonators: bool,
        include_fictional_czs: bool,
        metrics: ObservationFinder | None = None,
    ):
        super().__init__()
        # HACK: As of Qiskit 1.3, Target.__init__ does nothing with its args, instead Target.__new__
        # handles them.  IQMTarget.__init__ has different args than Target.__init__, hence we must either
        # (1) define IQMTarget.__new__ with matching args that calls Target.__new__, or
        # (2) make all the IQMTarget.__init__ args keyword-only with non-colliding names, and init
        # the necessary superclass attributes ourselves.
        # (1) seems to break pickling during concurrent transpilation using Qiskit's ``transpile``, so we use (2).
        self.qubit_properties = self._create_qubit_properties(architecture.qubits, metrics)

        # Using iqm_ as a prefix to avoid name clashes with the base class.
        self.iqm_dqa = architecture
        self.iqm_component_to_idx = component_to_idx
        self.iqm_idx_to_component = {v: k for k, v in component_to_idx.items()}
        self.iqm_include_resonators = include_resonators
        self.iqm_include_fictional_czs = include_fictional_czs
        self.iqm_metrics = metrics
        self._add_instructions_from_DQA()

    def _add_instructions_from_DQA(self):  # noqa: ANN202
        """Initializes the Target with instructions and properties that represent the
        dynamic quantum architecture :attr:`iqm_dqa`.

        """
        dqa = self.iqm_dqa
        # mapping from op name to all its allowed loci
        op_loci = {gate_name: gate_info.loci for gate_name, gate_info in dqa.gates.items()}

        def add_gate(gate: str) -> None:
            """Adds ``gate`` instructions to the Target."""
            props = {self.locus_to_idx(locus): self._create_gate_properties(gate, locus) for locus in op_loci[gate]}
            self.add_instruction(_QISKIT_IQM_GATE_MAP[gate], props)

        # identity gate does nothing and is removed in serialization, so we may as well allow it everywhere
        # Except if it is defined for the resonator, the graph is disconnected and the transpiler will fail.
        self.add_instruction(
            IGate(),
            {
                self.locus_to_idx((component,)): None
                for component in (dqa.components if self.iqm_include_resonators else dqa.qubits)
            },
        )

        # like barrier, delay is always available for all single-qubit loci
        self.add_instruction(Delay(0), {self.locus_to_idx((q,)): None for q in dqa.qubits})

        if "measure" in op_loci:
            add_gate("measure")

        if "prx" in op_loci:
            add_gate("prx")

        # HACK reset gate shares cc_prx loci for now, until reset is also in the DQA/metrics
        if "cc_prx" in op_loci:
            self.add_instruction(
                _QISKIT_IQM_GATE_MAP["reset"],
                {self.locus_to_idx(locus): None for locus in op_loci["cc_prx"]},
            )

        if self.iqm_include_resonators and "move" in op_loci:
            add_gate("move")

        if "cz" in op_loci:
            self._add_instructions_cz_gates(op_loci)

    def _add_instructions_cz_gates(self, op_loci: dict[str, Iterable[Locus]]) -> None:
        """Adds CZ gate instructions to the Qiskit Target.

        This method handles both "real" and "fictional" CZ gates based on the
        provided operation loci.

        1.  **Real CZ Gates**:
            A CZ gate for a locus is considered "real" if the locus
            exists in ``dynamic_quantum_architecture.gates['cz']``.
            A "real" CZ instruction is added if either the target is configured to
            include resonators (:attr:`iqm_include_resonators` is True) or if all
            the locus components are qubits.

        2.  **Fictional CZ Gates**: For STAR QPU architectures, a "fictional" CZ between
            two qubits that are not physically connected is defined
            iff there is a CZ gate from one qubit to a resonator, and a MOVE gate
            from the other qubit to the resonator.
            See :mod:`iqm.iqm_client.transpile`.

        Args:
            op_loci: Mapping of operation names (e.g., "cz", "move") to their available loci.

        """
        cz_props: dict[LocusIdx, InstructionProperties] = {}
        cz_loci = op_loci["cz"]

        # 1. Add real CZs
        for locus in cz_loci:
            if self.iqm_include_resonators or all(component in self.iqm_dqa.qubits for component in locus):
                locus_idx = self.locus_to_idx(locus)  # convert locus from IQM to Qiskit format
                cz_props[locus_idx] = self._create_gate_properties("cz", locus)

        # 2. Add fictional CZs if applicable
        if self.iqm_include_fictional_czs and "move" in op_loci:
            fictional_cz_loci = self._determine_fictional_cz_loci(op_loci["move"], cz_loci)
            for q1, q2, res in fictional_cz_loci:
                locus_idx = self.locus_to_idx((q1, q2))
                props = self._create_fictional_cz_properties(q1, q2, res)
                cz_props[locus_idx] = props
                cz_props[locus_idx[::-1]] = props  # fictional CZs are symmetric
                # TODO: if the fictional CZ can be implemented using both CZ(a,r) and MOVE(b,r), and
                # CZ(b,r) and MOVE(a,r), props will be only included for one implementation at random.
                # This is not necessarily the implementation the transpiler chooses!

        if cz_props:
            self.add_instruction(_QISKIT_IQM_GATE_MAP["cz"], cz_props)

    def _create_fictional_cz_properties(self, q1: str, q2: str, res: str) -> InstructionProperties:
        """Create ``InstructionProperties`` for a fictional CZ gate.

        A fictional CZ gate is comprised of the sequence MOVE(q2, res), CZ(q1, res), MOVE(q2, res).

        The duration and error for the fictional CZ gate are calculated as follows:

        - Duration: 2 * duration_move(q2, res) + duration_cz(q1, res)
        - Error: 1 - (1 - error_move(q2, res)))^2 * (1 - error_cz(q1, res))

        See https://arxiv.org/pdf/2503.10903 for more details.

        Args:
            q1: CZ qubit.
            q2: MOVE qubit.
            res: Shared computational resonator through which the fictional CZ gate is routed.

        Returns:
            Instruction properties for the fictional CZ gate.

        """
        if self.iqm_metrics is None:
            return None

        props_cz = self._create_gate_properties("cz", (q1, res))
        props_move = self._create_gate_properties("move", (q2, res))

        if props_cz is None or props_move is None:
            return None

        duration = None
        if props_move.duration is not None and props_cz.duration is not None:
            duration = 2 * props_move.duration + props_cz.duration

        error: float | None = None
        if props_move.error is not None and props_cz.error is not None:
            error = 1 - (1 - props_move.error) ** 2 * (1 - props_cz.error)

        return InstructionProperties(duration=duration, error=error)

    @staticmethod
    def _determine_fictional_cz_loci(move_loci: Iterable[Locus], cz_loci: Iterable[Locus]) -> list[Locus]:
        """Determine fictional CZ loci, i.e. pairs of qubits between which we can implement a CZ using MOVE gates and
        an intermediate computational resonator.

        Only applicable for the Star architecture.

        Args:
            move_loci: The loci for which a MOVE operation is defined.
            cz_loci: The loci for which a CZ operation is defined.

        Returns:
            Loci for fictional CZ gates, *including the resonator*. Each locus is given as ``(q1, q2, res)``.

        """
        fictional_cz_loci: list[Locus] = []
        for q1, res1 in cz_loci:
            for q2, res2 in move_loci:
                if res1 == res2 and q1 != q2:
                    # shared resonator, different qubits
                    fictional_cz_loci.append((q1, q2, res1))

        return fictional_cz_loci

    def locus_to_idx(self, locus: Locus) -> LocusIdx:
        """Map the given locus to use component indices instead of component names."""
        return tuple(self.iqm_component_to_idx[component] for component in locus)

    @staticmethod
    def _create_qubit_properties(
        qubits: Iterable[str], metrics: ObservationFinder | None
    ) -> list[QubitProperties] | None:
        """Creates qubit properties from the quality metrics."""
        if metrics is None:
            return None

        qubit_props = []
        t1_times, t2_times = metrics.get_coherence_times(qubits)
        for q in qubits:
            frequency = metrics.get_qubit_frequency(q)
            qubit_props.append(
                QubitProperties(
                    t1=t1_times.get(q),
                    t2=t2_times.get(q),
                    frequency=frequency,
                )
            )
        return qubit_props

    def _create_gate_properties(self, gate_name: str, locus: Locus) -> InstructionProperties | None:
        """Creates InstructionProperties for a single gate on a specific locus.

        Args:
            gate_name: Name of the IQM native operation to look up properties for (e.g., 'cz' or 'move').
            locus: Locus of the operation.

        Returns:
            Properties for the (default implementation of the) given gate at the given locus, or None if not available.

        """
        if self.iqm_metrics is None:
            return None

        # the properties are for the default implementation (other implementations may be available also)
        impl_name = self.iqm_dqa.gates[gate_name].get_default_implementation(locus)
        duration = self.iqm_metrics.get_gate_duration(gate_name, impl_name, locus)
        fidelity = self.iqm_metrics.get_gate_fidelity(gate_name, impl_name, locus)
        return InstructionProperties(duration=duration, error=None if fidelity is None else 1 - fidelity)

    @property
    def physical_qubits(self) -> list[str]:
        """Return the ordered list of physical qubits in the target."""
        # Overrides the property from the superclass to contain the correct information.
        return [self.iqm_idx_to_component[i] for i in range(self.num_qubits)]

    def restrict_to_qubits(self, qubits: list[int] | list[str]) -> IQMTarget:
        """Generated a restricted transpilation target from this Target that only contains the given qubits.

        Args:
            qubits: Qubits to restrict the target to. Can be either a list of qubit indices or qubit names.

        Returns:
            restricted target

        """
        qubits_str = [self.iqm_idx_to_component[q] if isinstance(q, int) else str(q) for q in qubits]
        return _restrict_dqa_to_qubits(
            self.iqm_dqa,
            qubits_str,
            self.iqm_include_resonators,
            self.iqm_include_fictional_czs,
            metrics=self.iqm_metrics,
        )


def _restrict_dqa_to_qubits(
    architecture: DynamicQuantumArchitecture,
    components: list[str],
    include_resonators: bool,
    include_fictional_czs: bool,
    metrics: ObservationFinder | None = None,
) -> IQMTarget:
    """Generated a restricted transpilation target that only contains the given QPU components.

    Args:
        architecture: The dynamic quantum architecture to restrict.
        components: QPU components to restrict the target to. Qubits first, then resonators.
        include_resonators: Whether to include computational resonators (and MOVE gates) in the target.
        include_fictional_czs: Whether to include fictional CZs that are not natively supported,
            but could be routed via MOVE.
        metrics: Optional calibration data and related quality metrics to improve the transpilation.

    Returns:
        restricted target

    """
    # include gate loci that only involve the given components
    c_set = set(components)
    new_gates = {}
    for gate_name, gate_info in architecture.gates.items():
        new_implementations = {}
        for implementation_name, implementation_info in gate_info.implementations.items():
            new_loci = tuple(locus for locus in implementation_info.loci if set(locus) <= c_set)
            if new_loci:
                new_implementations[implementation_name] = GateImplementationInfo(loci=new_loci)
        if new_implementations:
            new_gates[gate_name] = GateInfo(
                implementations=new_implementations,
                default_implementation=gate_info.default_implementation,
                override_default_implementation=gate_info.override_default_implementation,
            )
    new_arch = DynamicQuantumArchitecture(
        calibration_set_id=architecture.calibration_set_id,
        qubits=[q for q in architecture.qubits if q in c_set],
        computational_resonators=[r for r in architecture.computational_resonators if r in c_set],
        gates=new_gates,
    )
    return IQMTarget(
        architecture=new_arch,
        component_to_idx={name: idx for idx, name in enumerate(components)},
        include_resonators=include_resonators,
        include_fictional_czs=include_fictional_czs,
        metrics=metrics,
    )



# Copyright 2020–2025 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for serializing and deserializing quantum circuits between Cirq and IQM circuit formats."""

from math import pi

from cirq import Circuit, KeyCondition, NamedQid
from cirq.ops import (
    ClassicallyControlledOperation,
    CZPowGate,
    Gate,
    MeasurementGate,
    Operation,
    PhasedXPowGate,
    ResetChannel,
    XPowGate,
    YPowGate,
)
from iqm.cirq_iqm.iqm_gates import IQMMoveGate

import iqm.pulse
from iqm.pulse import CircuitOperation

# Mapping from IQM operation names to cirq operations
_IQM_CIRQ_OP_MAP: dict[str, tuple[type[Gate], ...]] = {
    # XPow and YPow kept for convenience, Cirq does not know how to decompose them into PhasedXPow
    # so we would have to add those rules...
    "prx": (PhasedXPowGate, XPowGate, YPowGate),
    "cc_prx": (),  # cc_prx is special, it does not directly map to any single Cirq op
    "cz": (CZPowGate,),
    "move": (IQMMoveGate,),
    "measure": (MeasurementGate,),
    "reset": (ResetChannel,),
}


def circuit_operation_to_operation(circuit_operation: CircuitOperation) -> Operation:
    """Convert an IQM circuit operation to a Cirq Operation.

    Args:
        circuit_operation: the IQM circuit_operation

    Returns:
        Operation: the converted operation

    Raises:
        OperationNotSupportedError When the circuit contains an unsupported operation.

    """
    if circuit_operation.name not in _IQM_CIRQ_OP_MAP:
        raise OperationNotSupportedError(f"Operation {circuit_operation.name} not supported.")

    qubits = [NamedQid(qubit, dimension=2) for qubit in circuit_operation.locus]

    if circuit_operation.name == "cc_prx":
        # special case
        args = {
            "exponent": circuit_operation.args["angle"] / pi,
            "phase_exponent": circuit_operation.args["phase"] / pi,
        }
        # ignore feedback_qubit, we currently only support 1-qubit KeyConditions so it can be
        # added in serialize_circuit
        return PhasedXPowGate(**args)(*qubits).with_classical_controls(circuit_operation.args["feedback_key"])

    cirq_op = _IQM_CIRQ_OP_MAP[circuit_operation.name][0]
    if circuit_operation.name == "prx":
        args = {
            "exponent": circuit_operation.args["angle"] / pi,
            "phase_exponent": circuit_operation.args["phase"] / pi,
        }
    elif circuit_operation.name == "measure":
        args = {"num_qubits": len(qubits), "key": circuit_operation.args["key"]}
        # ignore feedback_key, in Cirq it has to be the same as key
    else:
        # cz, move, reset have no args
        args = {}
    return cirq_op(**args)(*qubits)


class OperationNotSupportedError(RuntimeError):
    """Raised when a given operation is not supported by the IQM server."""


def operation_to_circuit_operation(operation: Operation) -> CircuitOperation:
    """Map a Cirq Operation to the IQM data transfer format.

    Assumes the circuit has been transpiled so that it only contains operations natively supported by the
    given IQM quantum architecture.

    Args:
        operation: a Cirq Operation

    Returns:
        CircuitOperation: the converted operation

    Raises:
        OperationNotSupportedError When the circuit contains an unsupported operation.

    """
    locus = tuple(qubit.name if isinstance(qubit, NamedQid) else str(qubit) for qubit in operation.qubits)
    if isinstance(operation.gate, (PhasedXPowGate, XPowGate, YPowGate)):
        return CircuitOperation(
            name="prx",
            locus=locus,
            args={"angle": operation.gate.exponent * pi, "phase": operation.gate.phase_exponent * pi},
        )
    if isinstance(operation.gate, MeasurementGate):
        if any(operation.gate.full_invert_mask()):
            raise OperationNotSupportedError("Invert mask not supported")

        return CircuitOperation(
            name="measure",
            locus=locus,
            args={"key": operation.gate.key},
        )
    if isinstance(operation.gate, CZPowGate):
        if operation.gate.exponent == 1.0:
            return CircuitOperation(
                name="cz",
                locus=locus,
                args={},
            )
        raise OperationNotSupportedError(
            f"CZPowGate exponent was {operation.gate.exponent}, but only 1 is natively supported."
        )

    if isinstance(operation.gate, IQMMoveGate):
        return CircuitOperation(
            name="move",
            locus=locus,
            args={},
        )

    if isinstance(operation.gate, ResetChannel):
        return CircuitOperation(name="reset", locus=locus)

    if isinstance(operation, ClassicallyControlledOperation):
        if len(operation._conditions) > 1:
            raise OperationNotSupportedError("Classically controlled gates can currently only have one condition.")
        if not isinstance(operation._conditions[0], KeyCondition):
            raise OperationNotSupportedError("Only KeyConditions are supported as classical controls.")
        if isinstance(operation._sub_operation.gate, (PhasedXPowGate, XPowGate, YPowGate)):
            return CircuitOperation(
                name="cc_prx",
                locus=locus,
                args={
                    "angle": operation._sub_operation.gate.exponent * pi,
                    "phase": operation._sub_operation.gate.phase_exponent * pi,
                    "feedback_qubit": "",
                    "feedback_key": str(operation._conditions[0]),
                },
            )
        raise OperationNotSupportedError(
            f"Classical control on the {operation._sub_operation.gate} gate is not supported."
        )

        # skipping feedback_qubit and feedback_key information until total circuit serialization

    raise OperationNotSupportedError(f"{type(operation.gate)} not natively supported.")


def serialize_circuit(circuit: Circuit) -> iqm.pulse.Circuit:
    """Serializes a quantum circuit into the IQM data transfer format.

    Args:
        circuit: quantum circuit to serialize

    Returns:
        data transfer object representing the circuit

    """
    total_ops_list = [op for moment in circuit for op in moment]
    cc_prx_support = any(isinstance(op, ClassicallyControlledOperation) for op in total_ops_list)
    instructions = tuple(map(operation_to_circuit_operation, total_ops_list))

    if cc_prx_support:
        mkey_to_measurement: dict[str, CircuitOperation] = {}
        for inst in instructions:
            if inst.name == "measure":
                if inst.args["key"] in mkey_to_measurement:
                    raise OperationNotSupportedError("Cannot use the same key for multiple measurements.")
                mkey_to_measurement[inst.args["key"]] = inst
            elif inst.name == "cc_prx":
                feedback_key = inst.args["feedback_key"]
                measurement = mkey_to_measurement.get(feedback_key)
                if measurement is None:
                    raise OperationNotSupportedError(
                        f"cc_prx has feedback_key {feedback_key}, but no measure operation with that key precedes it."
                    )
                if len(measurement.locus) != 1:
                    raise OperationNotSupportedError("cc_prx must depend on the measurement result of a single qubit.")
                inst.args["feedback_qubit"] = measurement.locus[0]
                measurement.args["feedback_key"] = feedback_key

    return iqm.pulse.Circuit(name="Serialized from Cirq", instructions=instructions, metadata=None)


def deserialize_circuit(circuit: iqm.pulse.Circuit) -> Circuit:
    """Deserializes a quantum circuit from the IQM data transfer format to a Cirq Circuit.

    Args:
        circuit: data transfer object representing the circuit

    Returns:
        quantum circuit

    """
    return Circuit(
        map(
            circuit_operation_to_operation,
            circuit.instructions,
        )
    )



#  ********************************************************************************
#  Copyright (c) 2019-2025 IQM Finland Oy.
#  All rights reserved. Confidential and proprietary.
#
#  Distribution or reproduction of any information contained herein
#  is prohibited without IQM Finland Oy’s prior written permission.
#  ********************************************************************************
"""Validation related helper functions for IQMClient."""

from collections.abc import Iterable
import itertools

from iqm.iqm_client.errors import CircuitValidationError
from iqm.iqm_client.models import (
    _SUPPORTED_OPERATIONS,
    CircuitBatch,
    DynamicQuantumArchitecture,
    MoveGateValidationMode,
    QIRCode,
)

from iqm.pulse import Circuit, CircuitOperation


def validate_qubit_mapping(
    architecture: DynamicQuantumArchitecture,
    circuits: CircuitBatch,
    qubit_mapping: dict[str, str] | None = None,
) -> None:
    """Validate the given qubit mapping.

    Args:
      architecture: Quantum architecture to check against.
      circuits: Circuits to be checked.
      qubit_mapping: Mapping of logical qubit names to physical qubit names.
          Can be set to ``None`` if all ``circuits`` already use physical qubit names.
          Note that the ``qubit_mapping`` is used for all ``circuits``.

    Raises:
        CircuitValidationError: There was something wrong with ``circuits``.

    """
    if qubit_mapping is None:
        return

    # check if qubit mapping is injective
    target_qubits = set(qubit_mapping.values())
    if not len(target_qubits) == len(qubit_mapping):
        raise CircuitValidationError("Multiple logical qubits map to the same physical qubit.")

    # check if qubit mapping covers all qubits in the circuits
    for i, circuit in enumerate(circuits):
        if isinstance(circuit, (QIRCode)):
            continue
        diff = circuit.all_locus_components() - set(qubit_mapping)
        if diff:
            raise CircuitValidationError(
                f"The qubits {diff} in circuit '{circuit.name}' at index {i} "
                f"are not found in the provided qubit mapping."
            )

    # check that each mapped qubit is defined in the quantum architecture
    for _logical, physical in qubit_mapping.items():
        if physical not in architecture.components:
            raise CircuitValidationError(f"Component {physical} not present in dynamic quantum architecture")


def validate_circuit_instructions(
    architecture: DynamicQuantumArchitecture,
    circuits: CircuitBatch,
    qubit_mapping: dict[str, str] | None = None,
    validate_moves: MoveGateValidationMode = MoveGateValidationMode.STRICT,
    *,
    must_close_sandwiches: bool = True,
) -> None:
    """Validate the given circuits against the given quantum architecture.

    Args:
      architecture: Quantum architecture to check against.
      circuits: Circuits to be checked.
      qubit_mapping: Mapping of logical qubit names to physical qubit names.
          Can be set to ``None`` if all ``circuits`` already use physical qubit names.
          Note that the ``qubit_mapping`` is used for all ``circuits``.
      validate_moves: Determines how MOVE gate validation works.
      must_close_sandwiches: Iff True, MOVE sandwiches cannot be left open when the circuit ends.

    Raises:
        CircuitValidationError: validation failed

    """
    for index, circuit in enumerate(circuits):
        if isinstance(circuit, QIRCode):
            continue

        measurement_keys: set[str] = set()
        for instr in circuit.instructions:
            validate_instruction(architecture, instr, qubit_mapping)
            # check measurement key uniqueness
            if instr.name in {"measure", "measurement"}:
                key = instr.args["key"]
                if key in measurement_keys:
                    raise CircuitValidationError(f"Circuit {index}: {instr!r} has a non-unique measurement key.")
                measurement_keys.add(key)
        validate_circuit_moves(
            architecture,
            circuit,
            qubit_mapping,
            validate_moves=validate_moves,
            must_close_sandwiches=must_close_sandwiches,
        )


def validate_instruction(
    architecture: DynamicQuantumArchitecture,
    instruction: CircuitOperation,
    qubit_mapping: dict[str, str] | None = None,
) -> None:
    """Validate an instruction against the dynamic quantum architecture.

    Checks that the instruction uses a valid implementation, and targets a valid locus.

    Args:
      architecture: Quantum architecture to check against.
      instruction: Instruction to check.
      qubit_mapping: Mapping of logical qubit names to physical qubit names.
          Can be set to ``None`` if ``instruction`` already uses physical qubit names.

    Raises:
        CircuitValidationError: validation failed

    """
    op_info = _SUPPORTED_OPERATIONS.get(instruction.name)
    if op_info is None:
        raise CircuitValidationError(f"Unknown quantum operation '{instruction.name}'.")

    # apply the qubit mapping if any
    mapped_qubits = tuple(qubit_mapping[q] for q in instruction.locus) if qubit_mapping else instruction.locus

    def check_locus_components(allowed_components: Iterable[str], msg: str) -> None:
        """Checks that the instruction locus consists of the allowed components only."""
        for q, mapped_q in zip(instruction.locus, mapped_qubits):
            if mapped_q not in allowed_components:
                raise CircuitValidationError(
                    f"{instruction!r}: Component {q} = {mapped_q} {msg}."
                    if qubit_mapping
                    else f"{instruction!r}: Component {q} {msg}."
                )

    impl_name = instruction.implementation or _SUPPORTED_OPERATIONS[
        instruction.name
    ].get_default_implementation_for_locus(instruction.locus)
    if (impl := op_info.implementations.get(impl_name)) is not None:
        if not impl.needs_calibration():
            # all QPU loci are allowed
            check_locus_components(architecture.components, msg="does not exist on the QPU")
            return

    gate_info = architecture.gates.get(instruction.name)
    if gate_info is None:
        raise CircuitValidationError(
            f"Operation '{instruction.name}' is not supported by the dynamic quantum architecture."
        )

    if instruction.implementation is not None:
        # specific implementation requested
        impl_info = gate_info.implementations.get(instruction.implementation)
        if impl_info is None:
            raise CircuitValidationError(
                f"Operation '{instruction.name}' implementation '{instruction.implementation}' "
                f"is not supported by the dynamic quantum architecture."
            )
        allowed_loci = impl_info.loci
        instruction_name = f"{instruction.name}.{instruction.implementation}"
    else:
        # any implementation is fine
        allowed_loci = gate_info.loci
        instruction_name = f"{instruction.name}"

    if op_info.factorizable:
        # Check that all the locus components are allowed by the architecture
        check_locus_components(
            {q for locus in allowed_loci for q in locus}, msg=f"is not allowed as locus for '{instruction_name}'"
        )
        return

    # Check that locus matches one of the allowed loci
    all_loci = (
        tuple(tuple(x) for locus in allowed_loci for x in itertools.permutations(locus))
        if op_info.symmetric
        else allowed_loci
    )
    if mapped_qubits not in all_loci:
        raise CircuitValidationError(
            f"{instruction.locus} = {tuple(mapped_qubits)} is not allowed as locus for '{instruction_name}'"
            if qubit_mapping
            else f"{instruction.locus} is not allowed as locus for '{instruction_name}'"
        )


def validate_circuit_moves(  # noqa: PLR0912
    architecture: DynamicQuantumArchitecture,
    circuit: Circuit,
    qubit_mapping: dict[str, str] | None = None,
    validate_moves: MoveGateValidationMode = MoveGateValidationMode.STRICT,
    *,
    must_close_sandwiches: bool = True,
) -> None:
    """Raise an error if the MOVE gates in the circuit are not valid in the given architecture.

    Args:
        architecture: Quantum architecture to check against.
        circuit: Quantum circuit to validate.
        qubit_mapping: Mapping of logical qubit names to physical qubit names.
            Can be set to ``None`` if the ``circuit`` already uses physical qubit names.
        validate_moves: Option for bypassing full or partial MOVE gate validation.
        must_close_sandwiches: Iff True, MOVE sandwiches cannot be left open when the circuit ends.

    Raises:
        CircuitValidationError: validation failed

    """
    if validate_moves == MoveGateValidationMode.NONE:
        return
    move_gate = "move"
    # Check if MOVE gates are allowed on this architecture
    if move_gate not in architecture.gates:
        if any(i.name == move_gate for i in circuit.instructions):
            raise CircuitValidationError("MOVE instruction is not supported by the given device architecture.")
        return

    # some gates are allowed in MOVE sandwiches
    allowed_gates = {"barrier"}
    if validate_moves == MoveGateValidationMode.ALLOW_PRX:
        allowed_gates.add("prx")

    all_resonators = set(architecture.computational_resonators)
    all_qubits = set(architecture.qubits)
    if qubit_mapping:
        reverse_mapping = {phys: log for log, phys in qubit_mapping.items()}
        all_resonators = {reverse_mapping[q] if q in reverse_mapping else q for q in all_resonators}
        all_qubits = {reverse_mapping[q] if q in reverse_mapping else q for q in all_qubits}

    # Mapping from resonator to the qubit whose state it holds. Resonators not in the map hold no qubit state.
    resonator_occupations: dict[str, str] = {}
    # Qubits whose states are currently moved to a resonator
    moved_qubits: set[str] = set()

    for inst in circuit.instructions:
        if inst.name == "move":
            qubit, resonator = inst.locus
            if not (qubit in all_qubits and resonator in all_resonators):
                raise CircuitValidationError(
                    f"MOVE instructions are only allowed between qubit and resonator, not {inst.locus}."
                )

            if (resonator_qubit := resonator_occupations.get(resonator)) is None:
                # Beginning MOVE: check that the qubit hasn't been moved to another resonator
                if qubit in moved_qubits:
                    raise CircuitValidationError(
                        f"MOVE instruction {inst.locus}: state of {qubit} is "
                        f"in another resonator: {resonator_occupations}."
                    )
                resonator_occupations[resonator] = qubit
                moved_qubits.add(qubit)
            else:
                # Ending MOVE: check that the qubit matches to the qubit that was moved to the resonator
                if resonator_qubit != qubit:
                    raise CircuitValidationError(
                        f"MOVE instruction {inst.locus} to an already occupied resonator: {resonator_occupations}."
                    )
                del resonator_occupations[resonator]
                moved_qubits.remove(qubit)
        elif moved_qubits:
            # Validate that moved qubits are not used during MOVE operations
            if inst.name not in allowed_gates:
                if overlap := set(inst.locus) & moved_qubits:
                    raise CircuitValidationError(
                        f"Instruction {inst.name} acts on {inst.locus} while the state(s) of {overlap} "
                        f"are in a resonator. Current resonator occupation: {resonator_occupations}."
                    )

    # Finally validate that all MOVE sandwiches have been ended before the circuit ends
    if must_close_sandwiches and resonator_occupations:
        raise CircuitValidationError(
            f"Circuit ends while qubit state(s) are still in a resonator: {resonator_occupations}."
        )
